/*
https://github.com/peterix/dfhack
Copyright (c) 2009-2011 Petr Mrázek (peterix@gmail.com)

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any
damages arising from the use of this software.

Permission is granted to anyone to use this software for any
purpose, including commercial applications, and to alter it and
redistribute it freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this
software in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and
must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution.
*/

#include "Internal.h"

#include <string>
#include <vector>
#include <map>
#include <set>
#include <cstdio>
#include <cstring>
#include <iterator>
#include <sstream>
using namespace std;

#include "dfhack/Error.h"
#include "dfhack/Process.h"
#include "dfhack/Core.h"
#include "dfhack/Console.h"
#include "dfhack/Module.h"
#include "dfhack/VersionInfoFactory.h"
#include "dfhack/PluginManager.h"
#include "ModuleFactory.h"
#include "dfhack/modules/Gui.h"

#include "dfhack/SDL_fakes/events.h"

#include <stdio.h>
#include <iomanip>
using namespace DFHack;

void cheap_tokenise(string const& input, vector<string> &output)
{
    istringstream str(input);
    istream_iterator<string> cur(str), end;
    output.assign(cur, end);
}

struct IODATA
{
    Core * core;
    PluginManager * plug_mgr;
};

// A thread function... for handling hotkeys. This is needed because
// all the plugin commands are expected to be run from foreign threads.
// Running them from one of the main DF threads will result in deadlock!
int fHKthread(void * iodata)
{
    Core * core = ((IODATA*) iodata)->core;
    PluginManager * plug_mgr = ((IODATA*) iodata)->plug_mgr;
    if(plug_mgr == 0 || core == 0)
    {
        cerr << "Hotkey thread has croaked." << endl;
        return 0;
    }
    while(1)
    {
        std::string stuff = core->getHotkeyCmd(); // waits on mutex!
        if(!stuff.empty())
        {
            vector <string> crap;
            plug_mgr->InvokeCommand(stuff, crap);
        }
    }
}

// A thread function... for the interactive console.
static bool flip0 = false;
int fIOthread(void * iodata)
{
    if(!flip0)
    {
        std::cerr << "Console from Thread " << SDL_ThreadID() << std::endl;
        flip0 = true;
    }
    Core * core = ((IODATA*) iodata)->core;
    PluginManager * plug_mgr = ((IODATA*) iodata)->plug_mgr;
    if(plug_mgr == 0 || core == 0)
    {
        dfout << "Something horrible happened to the plugin manager in Core's constructor..." << std::endl;
        return 0;
    }
    fprintf(dfout_C,"DFHack is ready. Have a nice day! Type in '?' or 'help' for help.\n");
    //dfterm <<  << endl;
    int clueless_counter = 0;
    while (true)
    {
        string command = "";
        dfout <<"[DFHack]# ";
        getline(cin, command);
        if (cin.eof())
        {
            command = "q";
            dfout << std::endl; // No newline from the user here!
        }
        if(command=="help" || command == "?")
        {
            dfout << "Available commands" << endl;
            dfout << "------------------" << endl;
            for(int i = 0; i < plug_mgr->size();i++)
            {
                const Plugin * plug = (plug_mgr->operator[](i));
                dfout << "Plugin " << plug->getName() << " :" << std::endl;
                for (int j = 0; j < plug->size();j++)
                {
                    const PluginCommand & pcmd = (plug->operator[](j));
                    dfout << setw(12) << pcmd.name << "| " << pcmd.description << endl;
                }
                dfout << endl;
            }
        }
        else if( command == "" )
        {
            clueless_counter++;
        }
        else
        {
            vector <string> parts;
            cheap_tokenise(command,parts);
            if(parts.size() == 0)
            {
                clueless_counter++;
            }
            else
            {
                string first = parts[0];
                parts.erase(parts.begin());
                command_result res = plug_mgr->InvokeCommand(first, parts);
                if(res == CR_NOT_IMPLEMENTED)
                {
                    dfout << "Invalid command." << endl;
                    clueless_counter ++;
                }
                else if(res == CR_FAILURE)
                {
                    dfout << "ERROR!" << endl;
                }
            }
        }
        if(clueless_counter == 3)
        {
            dfout << "Do 'help' or '?' for the list of available commands." << endl;
            clueless_counter = 0;
        }
    }
}

Core::Core()
{
    // init the console. This must be always the first step!
    con = 0;
    plug_mgr = 0;
    vif = 0;
    p = 0;
    errorstate = false;
    vinfo = 0;
    started = false;
    memset(&(s_mods), 0, sizeof(s_mods));

    // create mutex for syncing with interactive tasks
    AccessMutex = 0;
    // set up hotkey capture
    memset(hotkey_states,0,sizeof(hotkey_states));
    hotkey_set = false;
    HotkeyMutex = 0;
    HotkeyCond = 0;
};

static bool flip1 = 0;
bool Core::Init()
{
    // init the console. This must be always the first step!
    con = new Console();
    if(!flip1)
    {
        std::cerr << "Construct from Thread " << SDL_ThreadID() << std::endl;
        flip1 = true;
    }
    // find out what we are...
    vif = new DFHack::VersionInfoFactory("Memory.xml");
    p = new DFHack::Process(vif);
    if (!p->isIdentified())
    {
        dfout << "Couldn't identify this version of DF." << std::endl;
        errorstate = true;
        delete p;
        p = NULL;
        return false;
    }
    vinfo = p->getDescriptor();

    // create mutex for syncing with interactive tasks
    AccessMutex = SDL_CreateMutex();
    if(!AccessMutex)
    {
        dfout << "Mutex creation failed." << std::endl;
        errorstate = true;
        return false;
    }
    // lock mutex
    SDL_mutexP(AccessMutex);

    // create plugin manager
    plug_mgr = new PluginManager(this);
    if(!plug_mgr)
    {
        dfout << "Failed to create the Plugin Manager." << std::endl;
        errorstate = true;
        return false;
    }
    // look for all plugins, 
    // create IO thread
    IODATA *temp = new IODATA;
    temp->core = this;
    temp->plug_mgr = plug_mgr;
    SDL::Thread * IO = SDL_CreateThread(fIOthread, (void *) temp);
    // set up hotkey capture
    HotkeyMutex = SDL_CreateMutex();
    HotkeyCond = SDL_CreateCond();
    SDL::Thread * HK = SDL_CreateThread(fHKthread, (void *) temp);
    started = true;
    return true;
}
/// sets the current hotkey command
bool Core::setHotkeyCmd( std::string cmd )
{
    // access command
    SDL_mutexP(HotkeyMutex);
    {
        hotkey_set = true;
        hotkey_cmd = cmd;
        SDL_CondSignal(HotkeyCond);
    }
    SDL_mutexV(HotkeyMutex);
    return true;
}
/// removes the hotkey command and gives it to the caller thread
std::string Core::getHotkeyCmd( void )
{
    string returner;
    SDL_mutexP(HotkeyMutex);
    while ( ! hotkey_set )
    {
        SDL_CondWait(HotkeyCond, HotkeyMutex);
    }
    hotkey_set = false;
    returner = hotkey_cmd;
    hotkey_cmd.clear();
    SDL_mutexV(HotkeyMutex);
    return returner;
}


void Core::Suspend()
{
    SDL_mutexP(AccessMutex);
}

void Core::Resume()
{
    /*
    for(unsigned int i = 0 ; i < allModules.size(); i++)
    {
        allModules[i]->OnResume();
    }
    */
    SDL_mutexV(AccessMutex);
}

// should always be from simulation thread!
static bool flip2 = false;
int Core::Update()
{
    if(!started) Init();
    if(errorstate)
        return -1;
    if(!flip2)
    {
        std::cerr << "Update from Thread " << SDL_ThreadID() << std::endl;
        flip2 = true;
    }
    // notify all the plugins that a game tick is finished
    plug_mgr->OnUpdate();
    SDL_mutexV(AccessMutex);
        // other threads can claim the mutex here and use DFHack.
        // NO CODE SHOULD EVER BE PLACED HERE
    SDL_mutexP(AccessMutex);
    return 0;
};

int Core::Shutdown ( void )
{
    errorstate = 1;
    if(plug_mgr)
    {
        delete plug_mgr;
        plug_mgr = 0;
    }
    // invalidate all modules
    for(unsigned int i = 0 ; i < allModules.size(); i++)
    {
        delete allModules[i];
    }
    allModules.clear();
    memset(&(s_mods), 0, sizeof(s_mods));
    dfout << std::endl;
    // kill the console object
    if(con)
        delete con;
    con = 0;
    return -1;
}

static bool flip3 = 0;
int Core::SDL_Event(SDL::Event* ev, int orig_return)
{
    // do NOT process events before we are ready.
    if(!started) return orig_return;
    if(!flip3)
    {
        std::cerr << "Event from Thread " << SDL_ThreadID() << std::endl;
        flip3 = true;
    }
    if(!ev)
        return orig_return;
    if(ev && ev->type == SDL::ET_KEYDOWN || ev->type == SDL::ET_KEYUP)
    {
        SDL::KeyboardEvent * ke = (SDL::KeyboardEvent *)ev;
        bool shift = ke->ksym.mod & SDL::KMOD_SHIFT;
        // consuming F1 .. F8
        int idx = ke->ksym.sym - SDL::K_F1;
        if(idx < 0 || idx > 7)
            return orig_return;
        idx += 8*shift;
        // now we have the real index...
        if(ke->state == SDL::BTN_PRESSED && !hotkey_states[idx])
        {
            hotkey_states[idx] = 1;
            Gui * g = getGui();
            if(g->hotkeys && g->interface && g->menu_state)
            {
                t_viewscreen * ws = g->GetCurrentScreen();
                if(ws->getClassName() == "viewscreen_dwarfmodest" && *g->menu_state == 0x23)
                    return orig_return;
                else
                {
                    std::cerr << "Hotkey " << idx << " triggered. Thread " << SDL_ThreadID() << std::endl;
                    t_hotkey & hotkey = (*g->hotkeys)[idx];
                    setHotkeyCmd(hotkey.name);
                }
            }
        }
        else if(ke->state == SDL::BTN_RELEASED)
        {
            hotkey_states[idx] = 0;
        }
    }
    return orig_return;
    // do stuff with the events...
}

/*******************************************************************************
                                M O D U L E S
*******************************************************************************/

#define MODULE_GETTER(TYPE) \
TYPE * Core::get##TYPE() \
{ \
    if(errorstate) return NULL;\
    if(!s_mods.p##TYPE)\
    {\
        Module * mod = create##TYPE();\
        s_mods.p##TYPE = (TYPE *) mod;\
        allModules.push_back(mod);\
    }\
    return s_mods.p##TYPE;\
}

MODULE_GETTER(Creatures);
MODULE_GETTER(Engravings);
MODULE_GETTER(Maps);
MODULE_GETTER(Gui);
MODULE_GETTER(World);
MODULE_GETTER(Materials);
MODULE_GETTER(Items);
MODULE_GETTER(Translation);
MODULE_GETTER(Vegetation);
MODULE_GETTER(Buildings);
MODULE_GETTER(Constructions);